#include "DISPredictor.h"

using namespace DISPred;



DISPredictor* DISPredictor::Instance()
{
  static DISPredictor inst;
  return &inst;
}

void DISPredictor::Initialise(int my_argc, char **my_argv){


  // initialise cards
  InitCards();

  // read cards
  ReadCards(my_argc,my_argv);


  // Initialise root files
  my_file= new TFile(ControlCards::Instance()->fetchValueString("ROOTOutputFile").c_str(),"RECREATE");
  out_tree= new TTree("ReducedCrossSections","ReducedCrossSections");
  out_tree->Branch("Q2",&_Tq2,"Q2/D");
  out_tree->Branch("x",&_Tx,"x/D");
  out_tree->Branch("d",&_Td,"d/D");
  out_tree->Branch("u",&_Tu,"u/D");
  out_tree->Branch("s",&_Ts,"s/D");
  out_tree->Branch("c",&_Tc,"c/D");
  out_tree->Branch("b",&_Tb,"b/D");
  out_tree->Branch("dbar",&_Tdbar,"dbar/D");
  out_tree->Branch("ubar",&_Tubar,"ubar/D");
  out_tree->Branch("sbar",&_Tsbar,"sbar/D");
  out_tree->Branch("cbar",&_Tcbar,"cbar/D");
  out_tree->Branch("bbar",&_Tbbar,"bbar/D");
  out_tree->Branch("redsigma",&_Tredsigma,"redsigma/D");
  out_tree->Branch("ddiffsigma",&_Tddiffsigma,"ddiffsigma/D");
  out_tree->Branch("point",&_Tpoint,"point/i");
  _Tpoint=0;

  InitSingleDifferentials();

  // Calculate Centre-Of-Mass Energy
  CalculateCentreOfMassEnergy();

  // initialise PDFs
  InitPDF(ControlCards::Instance()->fetchValueInt("PDFSubSet"));

    if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"NLO")){
      if (ControlCards::Instance()->fetchValueString("PDFSetFileName").find("LHgrid")<ControlCards::Instance()->fetchValueString("PDFSetFileName").size()){
	qninit_();
	
	double mc2= pow(LHAPDF::getQMass(4),2);
	double mb2= pow(LHAPDF::getQMass(5),2);
	double mt2= pow(LHAPDF::getQMass(6),2);


	
	  

	
	std::cout << "NLO chosen, initialising qcdnum v16" << std::endl; 

	double up,dn,usea,dsea,str,chm,bot,glu,upv,dnv;
	double q0=7.0;
	Int_t   one=1;
	Int_t  Ix=409;
	Int_t  Iq=200; // was 61
	double xlim2=97e-8;
	double qlim1=29e-2;
	double qlim2=200e3;
	Int_t Iq2def=199;
	Int_t Ixdef=409;
	double Dxdef=4e-4;
	double Dqdef1=4.;
	double Dqdef2=200e3;
	
	
	// some heavy flavour stuff
	
	double Q2C=1.8225;
	double Q2B=18.49;
	double Q2CC=Q2C;
	double Q2BB=Q2B;
	
	double qsdt=4.*Q2C;
	double qsct=4.*Q2B;
	
	
	
	grxdef(Ix,xlim2);
	grqdef(Iq,qlim1,qlim2);
	double qs[3]={q0,Q2C,Q2B};
	double qmore=50.0;
	grqinp(qmore,1);
	grqinp(mc2,1);
	grqinp(mb2,1);
	qthres(mc2,mb2);

	
	
	
	
	
	
	double Mz= ControlCards::Instance()->fetchValueDouble("ZBosonMass");
	double alfas=LHAPDF::alphasPDF(Mz);
	qnrset("ALFQ0",Mz*Mz);
	qnrset("ALFAS",alfas);
	qnrset("AAAR2",1.);
	qnrset("BBBR2",0.);
	qnrset("AAM2L",1.);
	qnrset("BBM2L",0.);
	qnrset("AAM2H",1.);
	qnrset("BBM2H",0.);
	qnrset("MCSTF",sqrt(mc2));
	qnrset("MBSTF",sqrt(mb2));
	qnrset("MCALF",sqrt(mc2));
	qnrset("MBALF",sqrt(mb2));

	qnlset("WTF2C",true); 
	qnlset("W2STF",true); 
	qnlset("WTF2B",true) ;
	qnlset("CLOWQ",false) ;
	qnlset("WTFLC",true) ;
	qnlset("WTFLB",true) ;
	
	qniset("ORDER",2) ;
	


	
	qnfilw(0,0);      
	
       	Int_t Iq0=iqfromq(q0);
	Int_t Iqc=iqfromq(Q2C);
	Int_t Iqb=iqfromq(Q2B);
	
	
	qnbook(2,"UPLUS");
	qnbook(3,"DPLUS");
	qnbook(4,"SPLUS");
	qnbook(5,"CPLUS");
	qnbook(6,"BPLUS");
	qnbook(7,"UPVAL");
	qnbook(8,"DNVAL");

	
	
	for (int my_iq=1;my_iq<=Iq; my_iq++){
	  double nf=0.0;
	  double my_q=qfromiq(my_iq);
	  for( int mif=1;mif<7;mif++){
	   if (LHAPDF::xfx(0.01, sqrt(my_q), mif)>0.000000000001) nf=nf+1;
	  }
	  for (int mix=1;mix<=Ix;mix++){
	    double mx=xfromix(mix);
	    double singlet =0.0;
	    if (!strcmp(ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){
	      singlet=(LHAPDF::xfx(mx,sqrt(my_q), 1)+LHAPDF::xfx(mx,sqrt(my_q), 2)+LHAPDF::xfx(mx,sqrt(my_q), 3)+LHAPDF::xfx(mx,sqrt(my_q), 4)+LHAPDF::xfx(mx,sqrt(my_q), 5)+LHAPDF::xfx(mx,sqrt(my_q), -1)+LHAPDF::xfx(mx,sqrt(my_q), -2)+LHAPDF::xfx(mx,sqrt(my_q), -3)+LHAPDF::xfx(mx,sqrt(my_q), -4)+LHAPDF::xfx(mx,sqrt(my_q), -5));
	    } else{
	      singlet=(LHAPDF::xfx(mx,sqrt(my_q), 1)+LHAPDF::xfx(mx,sqrt(my_q), 2)+LHAPDF::xfx(mx,sqrt(my_q), 3)+LHAPDF::xfx(mx,sqrt(my_q), 4)+LHAPDF::xfx(mx,sqrt(my_q), 5)+LHAPDF::xfx(mx,sqrt(my_q), -1)+LHAPDF::xfx(mx,sqrt(my_q), -2)+LHAPDF::xfx(mx,sqrt(my_q), -3)+LHAPDF::xfx(mx,sqrt(my_q), -4)+LHAPDF::xfx(mx,sqrt(my_q), -5));
	    }
	    
	    qnpset("SINGL",mix,my_iq,singlet);
	    qnpset("UPLUS",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 2),0.0)+std::max(LHAPDF::xfx(mx, sqrt(my_q), -2),0.0)-(1.0/(nf))*singlet);
	    qnpset("DPLUS",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 1),0.0)+std::max(LHAPDF::xfx(mx, sqrt(my_q), -1),0.0)-(1.0/(nf))*singlet);
	    qnpset("SPLUS",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 3),0.0)+std::max(LHAPDF::xfx(mx, sqrt(my_q), -3),0.0)-(1.0/(nf))*singlet);
	    if (nf>3.01){
	      qnpset("CPLUS",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 4),0.0)+std::max(LHAPDF::xfx(mx, sqrt(my_q), -4),0.0)-(1.0/(nf))*singlet);
	    } else{
	      qnpset("CPLUS",mix,my_iq,0.0);
	    }
	    if (nf>4.01){
	      qnpset("BPLUS",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 5),0.0)+std::max(LHAPDF::xfx(mx, sqrt(my_q), -5),0.0)-(1.0/(nf))*singlet);
	    } else{
	      qnpset("BPLUS",mix,my_iq,0.0);
	    }
	    qnpset("GLUON",mix,my_iq,std::max(LHAPDF::xfx(mx,sqrt(my_q), 0),0.0));
	    qnpset("UPVAL",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 2),0.0)-std::max(LHAPDF::xfx(mx, sqrt(my_q), -2),0.0));
	    qnpset("DNVAL",mix,my_iq,std::max(LHAPDF::xfx(mx, sqrt(my_q), 1),0.0)-std::max(LHAPDF::xfx(mx, sqrt(my_q), -1),0.0));
	  }
	}
	int iflag=0;
      }
    }
    

  _redSigmas = new RedSigmaGrid();
  _dSigmaDQ2 = new   std::vector <DISPred::DiffSigmaPoint>;
  _dSigmaDX = new   std::vector <DISPred::DiffSigmaPoint>;
  _dSigmaDY = new   std::vector <DISPred::DiffSigmaPoint>;

  // Read in grids and bins
  if (!strcmp(ControlCards::Instance()->fetchValueString("ReducedCrossSection").c_str(),"ON")){
    if (!ReadRedSigmaGrid()){
      std::cerr << "Error reading reduced cross section points!" << std::endl;
    }
  }

  if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2").c_str(),"ON") && strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2Points").c_str(),"AUTO") ){
    if (!ReadDiffSigmaPoints(ControlCards::Instance()->fetchValueString("DSigmaDQ2Points").c_str(),_dSigmaDQ2)){
      std::cerr << "Error reading reduced cross section points!" << std::endl;
    }
  }

  if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDX").c_str(),"ON") && strcmp(ControlCards::Instance()->fetchValueString("DSigmaDXPoints").c_str(),"AUTO") ){
    if (!ReadDiffSigmaPoints(ControlCards::Instance()->fetchValueString("DSigmaDXPoints").c_str(),_dSigmaDX)){
      std::cerr << "Error reading reduced cross section points!" << std::endl;
    }
  }

  if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDY").c_str(),"ON") && strcmp(ControlCards::Instance()->fetchValueString("DSigmaDYPoints").c_str(),"AUTO") ){
    if (!ReadDiffSigmaPoints(ControlCards::Instance()->fetchValueString("DSigmaDYPoints").c_str(),_dSigmaDY)){
      std::cerr << "Error reading reduced cross section points!" << std::endl;
    }
  }

  


}


TH1D *  DISPredictor::MakeSingleDiffHisto( std::string fileName  , std::string hist_name, double *binedges ){

  TH1D *hist;

  std::cout << "Making " << hist_name << " from " << fileName << std::endl;

  int nbins=0;
  std::ifstream in(fileName.c_str());
  if (!in) exit(0);

  

  std::string s;
  std::string t;
  std::string c;  

  while (1) {
    getline(in,c);
    if (in.eof()) break;
    std::istringstream is(c);
    is >> t;
    if (t.length() == 0) continue;
    if (t[0] == '#') continue; 
    //    int ok = 0;
    binedges[nbins]=std::atof(t.c_str());
    std::cout << "nbins " <<  nbins << " binedge " << binedges[nbins] << std::endl;
    nbins++;
  }
  
  

    hist=new TH1D(hist_name.c_str(),hist_name.c_str(),(nbins-1),binedges);

  std::cout << "Testing histo" << std::endl;
  for (int i=1;i<=hist->GetNbinsX();i++){
    std::cout << "nbin " <<  i  << " binedge " << hist->GetBinLowEdge(i) << " " << hist->GetBinLowEdge(i)+hist->GetBinWidth(i) << std::endl;
  }



  return hist;
}

void DISPredictor::InitSingleDifferentials(){
  
    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDX").c_str(),"ON")){
      _dSigdX  =   MakeSingleDiffHisto(ControlCards::Instance()->fetchValueString("DSigmaDXBins"),"DSigmaDX",_xbinedges);
      _XdSigdX  =   MakeSingleDiffHisto(ControlCards::Instance()->fetchValueString("DSigmaDXBins"),"XDSigmaDX",_xbinedges);
      
    }else{
      _dSigdX=new TH1D("DSigmaDX","DSigmaDX",1,0,1);
      _XdSigdX=new TH1D("XDSigmaDX","XDSigmaDX",1,0,1);
    }

    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDY").c_str(),"ON")){
      _dSigdY  =   MakeSingleDiffHisto(ControlCards::Instance()->fetchValueString("DSigmaDYBins"),"DSigmaDY",_xbinedges);
      _YdSigdY  =   MakeSingleDiffHisto(ControlCards::Instance()->fetchValueString("DSigmaDYBins"),"YDSigmaDY",_xbinedges);
      
    }else{
      _dSigdY=new TH1D("DSigmaDY","DSigmaDY",1,0,1);
      _YdSigdY=new TH1D("YDSigmaDY","YDSigmaDY",1,0,1);
    }


    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2").c_str(),"ON")){
      _dSigdQ2  =   MakeSingleDiffHisto(ControlCards::Instance()->fetchValueString("DSigmaDQ2Bins"),"DSigmaDQ2",_q2binedges);
      _Q2dSigdQ2  =   MakeSingleDiffHisto(ControlCards::Instance()->fetchValueString("DSigmaDQ2Bins"),"Q2DSigmaDQ2",_q2binedges);
      
    }else{

      _dSigdQ2=new TH1D("DSigmaDQ2","DSigmaDQ2",1,200,5e4);
      _Q2dSigdQ2=new TH1D("Q2DSigmaDQ2","Q2DSigmaDQ2",1,200,5e4);
    }
}

void DISPredictor::CalculateCrossSections(){

  if (!strcmp(ControlCards::Instance()->fetchValueString("ReducedCrossSection").c_str(),"ON")){
    for (unsigned int i=0;i<_redSigmas->size();i++){
      if (!strcmp(ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){
	_redSigmas->at(i).SetRedSigma(CalculateCCReducedCrossSection(_redSigmas->at(i).X(), _redSigmas->at(i).Q2() ));				      _redSigmas->at(i).SetD2sDQ2Dx(CalculateCCPropagator(_redSigmas->at(i).Q2(),_redSigmas->at(i).X())*_redSigmas->at(i).RedSigma() );
	_Tpoint++;
	_Tq2=_redSigmas->at(i).Q2();
	_Tx=_redSigmas->at(i).X();
	_Td=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 1);
	_Tdbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -1);
	_Tu=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 2);
	_Tubar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -2);
	_Ts=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 3);
	_Tsbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -3);
	_Tc=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 4);
	_Tcbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -4);
	_Tb=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 5);
	_Tbbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -5);


	_Tredsigma=_redSigmas->at(i).RedSigma();
	_Tddiffsigma=_redSigmas->at(i).D2sDQ2Dx();



	out_tree->Fill();

      } else{
	_redSigmas->at(i).SetRedSigma(CalculateReducedCrossSection(_redSigmas->at(i).X(), _redSigmas->at(i).Q2()) );
	_redSigmas->at(i).SetD2sDQ2Dx(CalculatePropagator(_redSigmas->at(i).Q2(),_redSigmas->at(i).X())*_redSigmas->at(i).RedSigma() );
	_Tpoint++;
	_Tq2=_redSigmas->at(i).Q2();
	_Tx=_redSigmas->at(i).X();
	_Td=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 1);
	_Tdbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -1);
	_Tu=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 2);
	_Tubar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -2);
	_Ts=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 3);
	_Tsbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -3);
	_Tc=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 4);
	_Tcbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -4);
	_Tb=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), 5);
	_Tbbar=LHAPDF::xfx(_redSigmas->at(i).X(),sqrt(_redSigmas->at(i).Q2()), -5);
	_Tredsigma=_redSigmas->at(i).RedSigma();
	_Tddiffsigma=_redSigmas->at(i).D2sDQ2Dx();
	out_tree->Fill();
      }
    }
  }

    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2").c_str(),"ON")){
        for (int i=1; i<=_dSigdQ2->GetNbinsX();i++){
	  _dSigdQ2->SetBinContent(i, CalculateDSigmaDQ2(_dSigdQ2->GetBinLowEdge(i),_dSigdQ2->GetBinLowEdge(i)+_dSigdQ2->GetBinWidth(i))) ;
	  
	  if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2Points").c_str(),"AUTO")){
	    _Q2dSigdQ2->SetBinContent(i, CalculateQ2DSigmaDQ2(_dSigdQ2->GetBinLowEdge(i),_dSigdQ2->GetBinLowEdge(i)+_dSigdQ2->GetBinWidth(i))) ;
	    _dSigmaDQ2->push_back(DISPred::DiffSigmaPoint(_Q2dSigdQ2->GetBinContent(i)/_dSigdQ2->GetBinContent(i)));
	  }
	  
	}
	for (unsigned int i=0;i<_dSigmaDQ2->size();i++){
	  double binwidth=ControlCards::Instance()->fetchValueDouble("DiffBinPointScale")*_dSigdQ2->GetBinWidth(_dSigdQ2->FindBin(_dSigmaDQ2->at(i).Var()));
	  _dSigmaDQ2->at(i).SetDiffSigma(CalculateDSigmaDQ2(_dSigmaDQ2->at(i).Var()-(binwidth/2.0),_dSigmaDQ2->at(i).Var()+(binwidth/2.0)));
	  _q2points[i]=_dSigmaDQ2->at(i).Var();
	  _q2vals[i]=_dSigmaDQ2->at(i).DiffSigma();
	  _q2errs[i]=0.0;
	}
	_GdSigdQ2= new TGraphAsymmErrors(_dSigmaDQ2->size(),_q2points,_q2vals,_q2errs,_q2errs); my_file->Add(_GdSigdQ2);
	_GdSigdQ2->SetName("GraphDSigmaDQ2");
	_GdSigdQ2->SetTitle("GraphDSigmaDQ2");

    }
    
    
    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDX").c_str(),"ON")){
      
      for (int i=1; i<=_dSigdX->GetNbinsX();i++){
	_dSigdX->SetBinContent(i, CalculateDSigmaDX(_dSigdX->GetBinLowEdge(i),_dSigdX->GetBinLowEdge(i)+_dSigdX->GetBinWidth(i))) ;
	if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDXPoints").c_str(),"AUTO")){
	  _XdSigdX->SetBinContent(i, CalculateXDSigmaDX(_dSigdX->GetBinLowEdge(i),_dSigdX->GetBinLowEdge(i)+_dSigdX->GetBinWidth(i))) ;
	  _dSigmaDX->push_back(DISPred::DiffSigmaPoint(_XdSigdX->GetBinContent(i)/_dSigdX->GetBinContent(i)));	    
	}
      }
      for (unsigned int i=0;i<_dSigmaDX->size();i++){
	double binwidth=ControlCards::Instance()->fetchValueDouble("DiffBinPointScale")*_dSigdX->GetBinWidth(_dSigdX->FindBin(_dSigmaDX->at(i).Var()));
	_dSigmaDX->at(i).SetDiffSigma(CalculateDSigmaDX(_dSigmaDX->at(i).Var()-(binwidth/2.0),_dSigmaDX->at(i).Var()+(binwidth/2.0)));
	_xpoints[i]=_dSigmaDX->at(i).Var();
	_xvals[i]=_dSigmaDX->at(i).DiffSigma();
	_xerrs[i]=0.0;
      }    
      _GdSigdX= new TGraphAsymmErrors(_dSigmaDX->size(),_xpoints,_xvals,_xerrs,_xerrs); my_file->Add(_GdSigdX);
      _GdSigdX->SetName("GraphDSigmaDX");
      _GdSigdX->SetTitle("GraphDSigmaDX");

    }
    

    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDY").c_str(),"ON")){
      for (int i=1; i<=_dSigdY->GetNbinsX();i++){
	_dSigdY->SetBinContent(i, CalculateDSigmaDY(_dSigdY->GetBinLowEdge(i),_dSigdY->GetBinLowEdge(i)+_dSigdY->GetBinWidth(i))) ;
	if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDYPoints").c_str(),"AUTO")){
	  _YdSigdY->SetBinContent(i, CalculateYDSigmaDY(_dSigdY->GetBinLowEdge(i),_dSigdY->GetBinLowEdge(i)+_dSigdY->GetBinWidth(i))) ;
	  _dSigmaDY->push_back(DISPred::DiffSigmaPoint(_YdSigdY->GetBinContent(i)/_dSigdY->GetBinContent(i)));	    
	}
      }
      for (unsigned int i=0;i<_dSigmaDY->size();i++){
	double binwidth=ControlCards::Instance()->fetchValueDouble("DiffBinPointScale")*_dSigdY->GetBinWidth(_dSigdY->FindBin(_dSigmaDY->at(i).Var()));
	_dSigmaDY->at(i).SetDiffSigma(CalculateDSigmaDY(_dSigmaDY->at(i).Var()-(binwidth/2.0),_dSigmaDY->at(i).Var()+(binwidth/2.0)));
	_ypoints[i]=_dSigmaDY->at(i).Var();
	_yvals[i]=_dSigmaDY->at(i).DiffSigma();
	_yerrs[i]=0.0;
      }
      _GdSigdY= new TGraphAsymmErrors(_dSigmaDY->size(),_ypoints,_yvals,_yerrs,_yerrs);my_file->Add(_GdSigdY);
      _GdSigdY->SetName("GraphDSigmaDY");
      _GdSigdY->SetTitle("GraphDSigmaDY");
      
    }
    
    
}



void  DISPredictor::CalculateCentreOfMassEnergy(){
  std::cout<< std::endl;
  std::cout << "Calculating Centre-of-Mass Energy"  << std::endl;
  double Ep= ControlCards::Instance()->fetchValueDouble("EProton");
  double El= ControlCards::Instance()->fetchValueDouble("ELepton");
  double Mp=0.938272;
  double Me=0.511e-3;
  _s= pow((Ep+El),2) - pow(sqrt(Ep*Ep-Mp*Mp)-sqrt(El*El-Me*Me),2);
  std::cout << "Electron Beam Energy: " << El  << " GeV" << std::endl;
  std::cout << "Proton Beam Energy: " << Ep  << " GeV" << std::endl;
  std::cout << "Centre-of-Mass Energy: " << sqrt(_s)  << " GeV"<< std::endl;
  std::cout<< std::endl;
}


double DISPredictor::CalculatePropagator(double q2, double x){
  double y=q2/(_s*x);
  double Yplus=(1+pow((1-y),2));
  double alpha=ControlCards::Instance()->fetchValueDouble("AlphaEM");
  double conversion_factor= 0.3894e9;  // GeV^2 -> pb^2
  return conversion_factor*2*TMath::Pi()*Yplus*alpha*alpha/(x*(q2*q2));
}


double DISPredictor::CalculateCCPropagator(double q2, double x){
  double Gf=ControlCards::Instance()->fetchValueDouble("GFermi");
  double mw=ControlCards::Instance()->fetchValueDouble("WBosonMass");
  double conversion_factor= 0.3894e9;  // GeV^2 -> pb^2
  return Gf*Gf*conversion_factor*pow(mw*mw/(q2+mw*mw),2)/(2*TMath::Pi()*x);
}

double DISPredictor::FL(double x, double q2){
  if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"LO")){
    return FL_LO(x,q2);
  } else {
    return FL_NLO(x,q2);
  }
}

double DISPredictor::FL_LO(double x, double q2){
  // always 0.0!
  return 0.0;
}

double DISPredictor::FL_NLO(double x, double q2){
  double Mz= ControlCards::Instance()->fetchValueDouble("ZBosonMass");
  double sin2thetaw= ControlCards::Instance()->fetchValueDouble("Sin2ThetaW");

  double ve=ControlCards::Instance()->fetchValueDouble("CouplingVe");
  double vu=ControlCards::Instance()->fetchValueDouble("CouplingVu");
  double vd=ControlCards::Instance()->fetchValueDouble("CouplingVd");

  double ae=ControlCards::Instance()->fetchValueDouble("CouplingAe");
  double au=ControlCards::Instance()->fetchValueDouble("CouplingAu");
  double ad=ControlCards::Instance()->fetchValueDouble("CouplingAd");

  double pz= q2/((Mz*Mz+q2)*(4*(sin2thetaw)*(1-sin2thetaw)));

  double Aup=4.0/9.0 -2*pz*(2.0/3.0)*(vu)*(ve) + pz*pz*(ve*ve+ae*ae)*(vu*vu+au*au);
  double Adown=1.0/9.0 -2*pz*(-1.0/3.0)*(vd)*(ve) + pz*pz*(ve*ve+ae*ae)*(vd*vd+ad*ad);
  double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
  weights[0]=(Aup+2*Adown)/3.0;
  weights[1] = Aup;
  weights[2] = Adown;
  weights[3] = Adown;
  qnlinc(23,"NCF2",3,weights);
  weights[0]=(Aup+Adown)/2;
  weights[4]=Aup;
  qnlinc(23,"NCF2",4,weights);
  weights[0]=(2.*Aup+3.*Adown)/5.;
  weights[5]=Adown;
  qnlinc(23,"NCF2",5,weights);
  
  int iflag=0;
  return qstfxq("FL","NCF2",x,q2,iflag);
}

double DISPredictor::F2(double x, double q2){

  if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"LO")){
    return F2_LO(x,q2);
  } else {
    return F2_NLO(x,q2);
  }

}

double DISPredictor::F2_NLO(double x, double q2){

  double Mz= ControlCards::Instance()->fetchValueDouble("ZBosonMass");
  double sin2thetaw= ControlCards::Instance()->fetchValueDouble("Sin2ThetaW");

  double ve=ControlCards::Instance()->fetchValueDouble("CouplingVe");
  double vu=ControlCards::Instance()->fetchValueDouble("CouplingVu");
  double vd=ControlCards::Instance()->fetchValueDouble("CouplingVd");

  double ae=ControlCards::Instance()->fetchValueDouble("CouplingAe");
  double au=ControlCards::Instance()->fetchValueDouble("CouplingAu");
  double ad=ControlCards::Instance()->fetchValueDouble("CouplingAd");

  double pz= q2/((Mz*Mz+q2)*(4*(sin2thetaw)*(1-sin2thetaw)));

  double Aup=4.0/9.0 -2*pz*(2.0/3.0)*(vu)*(ve) + pz*pz*(ve*ve+ae*ae)*(vu*vu+au*au);
  double Adown=1.0/9.0 -2*pz*(-1.0/3.0)*(vd)*(ve) + pz*pz*(ve*ve+ae*ae)*(vd*vd+ad*ad);
  double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
  weights[0]=(Aup+2*Adown)/3.0;
  weights[1] = Aup;
  weights[2] = Adown;
  weights[3] = Adown;
  qnlinc(23,"NCF2",3,weights);
  weights[0]=(Aup+Adown)/2;
  weights[4]=Aup;
  qnlinc(23,"NCF2",4,weights);
  weights[0]=(2.*Aup+3.*Adown)/5.;
  weights[5]=Adown;
  qnlinc(23,"NCF2",5,weights);
  int iflag=0;
  return qstfxq("F2","NCF2",x,q2,iflag);
}

double DISPredictor::F2_LO(double x, double q2){
  double Mz= ControlCards::Instance()->fetchValueDouble("ZBosonMass");
  double sin2thetaw= ControlCards::Instance()->fetchValueDouble("Sin2ThetaW");

  double ve=ControlCards::Instance()->fetchValueDouble("CouplingVe");
  double vu=ControlCards::Instance()->fetchValueDouble("CouplingVu");
  double vd=ControlCards::Instance()->fetchValueDouble("CouplingVd");

  double ae=ControlCards::Instance()->fetchValueDouble("CouplingAe");
  double au=ControlCards::Instance()->fetchValueDouble("CouplingAu");
  double ad=ControlCards::Instance()->fetchValueDouble("CouplingAd");

  double pz= q2/((Mz*Mz+q2)*(4*(sin2thetaw)*(1-sin2thetaw)));

  double Aup=4.0/9.0 -2*pz*(2.0/3.0)*(vu)*(ve) + pz*pz*(ve*ve+ae*ae)*(vu*vu+au*au);
  double Adown=1.0/9.0 -2*pz*(-1.0/3.0)*(vd)*(ve) + pz*pz*(ve*ve+ae*ae)*(vd*vd+ad*ad);


  double dterm=Adown*(LHAPDF::xfx(x, sqrt(q2), 1) + LHAPDF::xfx(x, sqrt(q2), -1));
  double uterm=Aup*(LHAPDF::xfx(x, sqrt(q2), 2) + LHAPDF::xfx(x, sqrt(q2), -2));  
  double sterm=Adown*(LHAPDF::xfx(x, sqrt(q2), 3) + LHAPDF::xfx(x, sqrt(q2), -3));  
  double cterm=Aup*(LHAPDF::xfx(x, sqrt(q2), 4) + LHAPDF::xfx(x, sqrt(q2), -4));  
  double bterm=Adown*(LHAPDF::xfx(x, sqrt(q2), 5) + LHAPDF::xfx(x, sqrt(q2), -5));  
  double tterm=Aup*(LHAPDF::xfx(x, sqrt(q2), 6) + LHAPDF::xfx(x, sqrt(q2), -6));  
  double sum=uterm+dterm+sterm+cterm+bterm+tterm;
  return sum;
}

double DISPredictor::XF3(double x, double q2){
  if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"LO")){
    return XF3_LO(x,q2);
  } else {
    return XF3_NLO(x,q2);
  }
}

double DISPredictor::XF3_LO(double x, double q2){
  double Mz= ControlCards::Instance()->fetchValueDouble("ZBosonMass");
  double sin2thetaw= ControlCards::Instance()->fetchValueDouble("Sin2ThetaW");

  double ve=ControlCards::Instance()->fetchValueDouble("CouplingVe");
  double vu=ControlCards::Instance()->fetchValueDouble("CouplingVu");
  double vd=ControlCards::Instance()->fetchValueDouble("CouplingVd");

  double ae=ControlCards::Instance()->fetchValueDouble("CouplingAe");
  double au=ControlCards::Instance()->fetchValueDouble("CouplingAu");
  double ad=ControlCards::Instance()->fetchValueDouble("CouplingAd");

  double pz= q2/((Mz*Mz+q2)*(4*(sin2thetaw)*(1-sin2thetaw)));

  double Bup=-2*(2.0/3.0)*au*ae*pz+4*au*ae*vu*ve*pz*pz;
  double Bdown=-2*(-1.0/3.0)*ad*ae*pz+4*ad*ae*vd*ve*pz*pz;
  double dterm=Bdown*(LHAPDF::xfx(x, sqrt(q2), 1) - LHAPDF::xfx(x, sqrt(q2), -1));
  double uterm=Bup*(LHAPDF::xfx(x, sqrt(q2), 2) - LHAPDF::xfx(x, sqrt(q2), -2));  
  double sterm=Bdown*(LHAPDF::xfx(x, sqrt(q2), 3) - LHAPDF::xfx(x, sqrt(q2), -3));  
  double cterm=Bup*(LHAPDF::xfx(x, sqrt(q2), 4) - LHAPDF::xfx(x, sqrt(q2), -4));  
  double bterm=Bdown*(LHAPDF::xfx(x, sqrt(q2), 5) - LHAPDF::xfx(x, sqrt(q2), -5));  
  double tterm=Bup*(LHAPDF::xfx(x, sqrt(q2), 6) - LHAPDF::xfx(x, sqrt(q2), -6));  
  double sum=(uterm+dterm+sterm+cterm+bterm+tterm);
  return sum;

}

double DISPredictor::XF3_NLO(double x, double q2){
  double Mz= ControlCards::Instance()->fetchValueDouble("ZBosonMass");
  double sin2thetaw= ControlCards::Instance()->fetchValueDouble("Sin2ThetaW");
  
  double ve=ControlCards::Instance()->fetchValueDouble("CouplingVe");
  double vu=ControlCards::Instance()->fetchValueDouble("CouplingVu");
  double vd=ControlCards::Instance()->fetchValueDouble("CouplingVd");

  double ae=ControlCards::Instance()->fetchValueDouble("CouplingAe");
  double au=ControlCards::Instance()->fetchValueDouble("CouplingAu");
  double ad=ControlCards::Instance()->fetchValueDouble("CouplingAd");
  
  double pz= q2/((Mz*Mz+q2)*(4*(sin2thetaw)*(1-sin2thetaw)));

  double Bup=-2*(2.0/3.0)*au*ae*pz+4*au*ae*vu*ve*pz*pz;
  double Bdown=-2*(-1.0/3.0)*ad*ae*pz+4*ad*ae*vd*ve*pz*pz;

  double Aup=4.0/9.0 -2*pz*(2.0/3.0)*(vu)*(ve) + pz*pz*(ve*ve+ae*ae)*(vu*vu+au*au);
  double Adown=1.0/9.0 -2*pz*(-1.0/3.0)*(vd)*(ve) + pz*pz*(ve*ve+ae*ae)*(vd*vd+ad*ad);

  double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[6]= Bup;
  weights[7]= Bdown;
  
  qnlinc(25,"NCF3",3,weights);
  qnlinc(25,"NCF3",4,weights);
  qnlinc(25,"NCF3",5,weights);

  int iflag=0;
  return qstfxq("XF3","NCF3",x,q2,iflag);
}


double DISPredictor::FLCC(double x, double q2){
  if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"LO")){
    return FLCC_LO(x,q2);
  } else {
    int iflag=0;
    return FLCC_NLO(x,q2);
  }
}

double DISPredictor::FLCC_LO(double x, double q2){

  
    return 0.0;

}

double DISPredictor::FLCC_NLO(double x, double q2){

  if (ControlCards::Instance()->fetchValueInt("ElectronCharge")>0){
    double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[0]=0.5;
    weights[6]=-0.5;
    weights[7]=0.5;  
    qnlinc(12,"CCEP2",3,weights);
    qnlinc(12,"CCEP2",4,weights);
    // actually you cannot just leave 5 the same as 4 for ccep2 because the b
    // will get added in by the weights(1), you have to get rid of it explicitly
    //    /*
    weights[0]=0.4;
    weights[1]=0.5;
    weights[2]=0.5;
    weights[3]=0.5;
    weights[4]=0.5;
    //    */
    qnlinc(12,"CCEP2",5,weights);
    
    int iflag=0;
    return qstfxq("FL","CCEP2",x,q2,iflag);
  } else {
    double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[0]=0.5;
    weights[6]=0.5;
    weights[7]=-0.5;  
    qnlinc(13,"CCEM2",3,weights);
    qnlinc(13,"CCEM2",4,weights);
    weights[0]=0.4;
    weights[1]=0.5;
    weights[2]=0.5;
    weights[3]=0.5;
    weights[4]=0.5;
    qnlinc(13,"CCEM2",5,weights);
    
    int iflag=0;
    return qstfxq("FL","CCEM2",x,q2,iflag);
  }





    return 0.0;

}

double DISPredictor::F2CC(double x, double q2){

  if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"LO")){
    return F2CC_LO(x,q2);
  } else {
    return F2CC_NLO(x,q2);
  }

}

double DISPredictor::F2CC_LO(double x, double q2){

  double m_t=ControlCards::Instance()->fetchValueDouble("TopMass");
  double m_b=ControlCards::Instance()->fetchValueDouble("BottomMass");
  double Vcb=ControlCards::Instance()->fetchValueDouble("Vcb");
  double Vub=ControlCards::Instance()->fetchValueDouble("Vub");

  if (ControlCards::Instance()->fetchValueInt("ElectronCharge")>0){
    double dterm=1*(LHAPDF::xfx(x, sqrt(q2), 1));
    double uterm=1*(LHAPDF::xfx(x, sqrt(q2), -2));  
    double sterm=1*(LHAPDF::xfx(x, sqrt(q2), 3));  
    double cterm=1*(LHAPDF::xfx(x, sqrt(q2), -4));  
    double bterm=1*(LHAPDF::xfx(x, sqrt(q2), 5));  
    double tterm=1*(LHAPDF::xfx(x, sqrt(q2), -6)); 
    if (q2<pow((m_t-m_b),2)){
      bterm=sqrt(Vub*Vub+Vcb*Vcb)*bterm;
    } 
    return(dterm+uterm+sterm+cterm+bterm+tterm);
  } else {
    double dterm=1*(LHAPDF::xfx(x, sqrt(q2), -1));
    double uterm=1*(LHAPDF::xfx(x, sqrt(q2), 2));  
    double sterm=1*(LHAPDF::xfx(x, sqrt(q2), -3));  
    double cterm=1*(LHAPDF::xfx(x, sqrt(q2), 4));  
    double bterm=1*(LHAPDF::xfx(x, sqrt(q2), -5));  
    double tterm=1*(LHAPDF::xfx(x, sqrt(q2), 6));  
    if (q2<pow((m_t-m_b),2)){
      bterm=sqrt(Vub*Vub+Vcb*Vcb)*bterm;
    }
    return(dterm+uterm+sterm+cterm+bterm+tterm);
  }

  return(-999.0);
}


double DISPredictor::F2CC_NLO(double x, double q2){

  if (ControlCards::Instance()->fetchValueInt("ElectronCharge")>0){
    double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[0]=0.5;
    weights[6]=-0.5;
    weights[7]=0.5;  
    qnlinc(12,"CCEP2",3,weights);
    qnlinc(12,"CCEP2",4,weights);
    // actually you cannot just leave 5 the same as 4 for ccep2 because the b
    // will get added in by the weights(1), you have to get rid of it explicitly
    // 
    weights[0]=0.4;
    weights[1]=0.5;
    weights[2]=0.5;
    weights[3]=0.5;
    weights[4]=0.5;
    // end lines not to be used
    qnlinc(12,"CCEP2",5,weights);

    int iflag=0;
    return qstfxq("F2","CCEP2",x,q2,iflag);
  } else {
    double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[0]=0.5;
    weights[6]=0.5;
    weights[7]=-0.5;  
    qnlinc(13,"CCEM2",3,weights);
    qnlinc(13,"CCEM2",4,weights);
    weights[0]=0.4;
    weights[1]=0.5;
    weights[2]=0.5;
    weights[3]=0.5;
    weights[4]=0.5;
    qnlinc(13,"CCEM2",5,weights);
    
    int iflag=0;
    return qstfxq("F2","CCEM2",x,q2,iflag);
  }
  return(0.0);
}

double DISPredictor::XF3CC(double x, double q2){
  if (!strcmp(ControlCards::Instance()->fetchValueString("QCDCalculationLevel").c_str(),"LO")){
    return XF3CC_LO(x,q2);
  } else {
    return XF3CC_NLO(x,q2);
  }
}

double DISPredictor::XF3CC_NLO(double x, double q2){
  double cos2thc= 1-ControlCards::Instance()->fetchValueDouble("Sin2ThetaC");
  double sin2thc=ControlCards::Instance()->fetchValueDouble("Sin2ThetaC");
  
  if (ControlCards::Instance()->fetchValueInt("ElectronCharge")>0){
    double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[1]=-0.5;
    weights[2]=0.5*cos2thc;
    weights[3]=0.5*sin2thc;
    weights[0]=0.;
    weights[6]=0.5;
    weights[7]=0.5*cos2thc;
    qnlinc(14,"CCEP3",3,weights);
    weights[1]=-0.5;
    weights[2]=0.5;
    weights[3]=0.5;
    weights[0]=0.;
    weights[4]=-0.5;
    weights[6]=0.5;
    weights[7]=0.5;
    qnlinc(14,"CCEP3",4,weights);
    // Leave b->t since it is mostly mass suppressed
    qnlinc(14,"CCEP3",5,weights);
    int iflag=0;
    return qstfxq("XF3","CCEP3",x,q2,iflag);
  } else {
    double weights[20]={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
    weights[1]=0.5;
    weights[2]=-0.5*cos2thc;
    weights[3]=-0.5*sin2thc;
    weights[0]=0.;
    weights[6]=0.5;
    weights[7]=0.5*cos2thc;
    qnlinc(15,"CCEM3",3,weights);
    weights[1]=0.5;
    weights[2]=-0.5;
    weights[3]=-0.5;
    weights[0]=0.;
    weights[4]=0.5;
    weights[6]=0.5;
    weights[7]=0.5;
    qnlinc(15,"CCEM3",4,weights);
    // Leave b->t since it is mostly mass suppressed
    qnlinc(15,"CCEM3",5,weights);
    int iflag=0;
    return qstfxq("XF3","CCEM3",x,q2,iflag);
  }

  return 0.0;
}

double DISPredictor::XF3CC_LO(double x, double q2){
  double m_t=ControlCards::Instance()->fetchValueDouble("TopMass");
  double m_b=ControlCards::Instance()->fetchValueDouble("BottomMass");
  double Vcb=ControlCards::Instance()->fetchValueDouble("Vcb");
  double Vub=ControlCards::Instance()->fetchValueDouble("Vub");

  if (ControlCards::Instance()->fetchValueInt("ElectronCharge")>0){
    double dterm=1*(LHAPDF::xfx(x, sqrt(q2), 1));
    double uterm=-1*(LHAPDF::xfx(x, sqrt(q2), -2));  
    double sterm=1*(LHAPDF::xfx(x, sqrt(q2), 3));  
    double cterm=-1*(LHAPDF::xfx(x, sqrt(q2), -4));  
    double bterm=1*(LHAPDF::xfx(x, sqrt(q2), 5)); 
    
    double tterm=-1*(LHAPDF::xfx(x, sqrt(q2), -6));  
    // don't include b to t term if Q^2 insufficient 
    if (q2<pow((m_t-m_b),2)){
      bterm=sqrt(Vub*Vub+Vcb*Vcb)*bterm;
    }
    return(dterm+uterm+sterm+cterm+bterm+tterm);
  } else {
    double dterm=-1*(LHAPDF::xfx(x, sqrt(q2), -1));
    double uterm=1*(LHAPDF::xfx(x, sqrt(q2), 2));  
    double sterm=-1*(LHAPDF::xfx(x, sqrt(q2), -3));  
    double cterm=1*(LHAPDF::xfx(x, sqrt(q2), 4));  
    double bterm=-1*(LHAPDF::xfx(x, sqrt(q2), -5));  
    double tterm=1*(LHAPDF::xfx(x, sqrt(q2), 6));  
    if (q2<pow((m_t-m_b),2)){
      bterm=sqrt(Vub*Vub+Vcb*Vcb)*bterm;
    }
    return(dterm+uterm+sterm+cterm+bterm+tterm);
  } 

  return -999.0;
}

double DISPredictor::CalculateCCReducedCrossSection(double x, double q2){
  
  double y=q2/(_s*x);
  double Yplus=(1+pow((1-y),2));
  double Yminus=(1-pow((1-y),2));

  double flterm=-0.5*y*y*FLCC(x,q2);
  double f2term=0.5*Yplus*F2CC(x,q2);
  double f3term=-0.5*ControlCards::Instance()->fetchValueInt("ElectronCharge")*Yminus*XF3CC(x,q2);

  
  return ((flterm+f2term+f3term));
}

double DISPredictor::CalculateReducedCrossSection(double x, double q2){
  

  double y=q2/(_s*x);
  double Yplus=(1+pow((1-y),2));
  double Yminus=(1-pow((1-y),2));

  double flterm=-y*y*FL(x,q2)/Yplus;
  double f2term=F2(x,q2);
  double f3term=-1*ControlCards::Instance()->fetchValueInt("ElectronCharge")*Yminus*XF3(x,q2)/Yplus;
  return (f2term+f3term+flterm);
}



void DISPredictor::InitCards(){

  // init cards
  ControlCards::Instance()->AddCardInt("ElectronCharge",1);
  ControlCards::Instance()->AddCardInt("VegasCalls",50000); // Calls to Vegas for numerical integration
  ControlCards::Instance()->AddCardString("PDFSetFileName","ZEUS2005_ZJ.LHpdf");
  ControlCards::Instance()->AddCardString("FlavourScheme","ZMVFN"); // Can be ZMVFN or RTVFN
  ControlCards::Instance()->AddCardInt("PDFSubSet",0);
  
  ControlCards::Instance()->AddCardString("QCDCalculationLevel","LO"); // can be NLO on NNLO in principle!

  ControlCards::Instance()->AddCardString("DISProcess","NC"); // CC/NC

  ControlCards::Instance()->AddCardDouble("ZBosonMass",91.1876); // Configurable!
  ControlCards::Instance()->AddCardDouble("WBosonMass",80.398); // Configurable!
  ControlCards::Instance()->AddCardDouble("AlphaEM",7.297352570e-03); // Configurable!
  ControlCards::Instance()->AddCardDouble("GFermi",1.16637e-05); // Configurable!
  ControlCards::Instance()->AddCardDouble("TopMass",171.2); // Configurable!
  ControlCards::Instance()->AddCardDouble("BottomMass",4.20); // Configurable!

  ControlCards::Instance()->AddCardDouble("Vub",41.2e-3); // Configurable!
  ControlCards::Instance()->AddCardDouble("Vcb",3.93e-3); // Configurable!

  ControlCards::Instance()->AddCardDouble("Sin2ThetaW",0.22308); // sin^2 theta_W Configurable!
  ControlCards::Instance()->AddCardDouble("Sin2ThetaC",0.05); // sin^2 theta_C Configurable!
  ControlCards::Instance()->AddCardDouble("CouplingVu",0.203); // SM 0.5 -4*sin2thetaw/3
  ControlCards::Instance()->AddCardDouble("CouplingVd",-0.351); // SM -0.5 +2*sin2thetaw/3
  ControlCards::Instance()->AddCardDouble("CouplingVe",-0.00538); // -0.5+2*sin2thetaw

  ControlCards::Instance()->AddCardDouble("CouplingAu",0.5);
  ControlCards::Instance()->AddCardDouble("CouplingAd",-0.5);
  ControlCards::Instance()->AddCardDouble("CouplingAe",-0.5);



  ControlCards::Instance()->AddCardString("ReducedCrossSection","OFF");
  ControlCards::Instance()->AddCardString("ReducedCrossSectionBins","q2xpoints.dat");

  ControlCards::Instance()->AddCardDouble("DiffBinPointScale",1e-6);
  ControlCards::Instance()->AddCardString("D2SigmaDQ2DX","OFF");

  ControlCards::Instance()->AddCardString("DSigmaDQ2","OFF");
  ControlCards::Instance()->AddCardString("DSigmaDQ2Bins","q2bins.dat");
  ControlCards::Instance()->AddCardString("DSigmaDQ2Points","AUTO");
  
  ControlCards::Instance()->AddCardString("DSigmaDX","OFF");
  ControlCards::Instance()->AddCardString("DSigmaDXBins","xbins.dat");
  ControlCards::Instance()->AddCardString("DSigmaDXPoints","AUTO");

  ControlCards::Instance()->AddCardString("DSigmaDY","OFF");
  ControlCards::Instance()->AddCardString("DSigmaDYBins","ybins.dat");
  ControlCards::Instance()->AddCardString("DSigmaDYPoints","AUTO");

  ControlCards::Instance()->AddCardDouble("Q2Min",0.0);
  ControlCards::Instance()->AddCardDouble("Q2Max",100000.0);

  ControlCards::Instance()->AddCardDouble("XMin",0.0);
  ControlCards::Instance()->AddCardDouble("XMax",1.0);

  ControlCards::Instance()->AddCardDouble("YMin",0.0);
  ControlCards::Instance()->AddCardDouble("YMax",1.0);


  ControlCards::Instance()->AddCardDouble("ELepton",27.5);
  ControlCards::Instance()->AddCardDouble("EProton",920.0);

  
  ControlCards::Instance()->AddCardString("ROOTOutputFile","DISPredOut.root");
  ControlCards::Instance()->AddCardString("EndCards","");

}

void DISPredictor::InitPDF(int subset){

  char name[160];
  strcpy(name,   ControlCards::Instance()->fetchValueString("PDFSetFileName").c_str());
  LHAPDF::initPDFSet(name,subset);

}


void DISPredictor::ReadCards(int my_argc, char **my_argv){

  if (my_argc>1){
    TString cardfile(my_argv[1]);
    if (ControlCards::Instance()->readKeys(cardfile) == 0) {
      std::stringstream ss1; ss1 << " data card file " << cardfile << " not found => using default values";
      std::cout << "WARNING: " << ss1.str()  << std::endl;
    }
  } else {
    std::stringstream ss1; ss1 << "usage:  Analysis <card name>.cards ";
    std::cout << "WARNING: " << ss1.str()   << std::endl;
    exit(-1);
  }
  

}

int DISPredictor::ReadRedSigmaGrid(){
  TString cardfile(ControlCards::Instance()->fetchValueString("ReducedCrossSectionBins").c_str());
  std::cout<< "ReadRedSigmaGrid: " << cardfile << std::endl;

  std::ifstream in(cardfile);
  if (!in) return 0;
  
  std::string s;
  std::string t;
  std::string u;
  std::string c;

  while (1) {
    getline(in,c);
    if (in.eof()) break;
    std::istringstream is(c);
    is >> t;
    if (t.length() == 0) continue;
    if (t[0] == '#') continue;


    is >> u ;

    RedSigmaPoint my_red=RedSigmaPoint(std::atof(t.c_str()),std::atof(u.c_str()));  
    _redSigmas->push_back(my_red);
    
  }
  
  return 1;
}



int DISPredictor::ReadDiffSigmaPoints(TString cardfile, std::vector <DISPred::DiffSigmaPoint> *points ){
  std::cout<< "ReadDiffSigmaPoint: " << cardfile << std::endl;

  std::ifstream in(cardfile);
  if (!in) return 0;
  
  std::string s;
  std::string t;
  std::string c;

  while (1) {
    getline(in,c);
    if (in.eof()) break;
    std::istringstream is(c);
    is >> t;
    if (t.length() == 0) continue;
    if (t[0] == '#') continue;



    DiffSigmaPoint my_red=DiffSigmaPoint(std::atof(t.c_str()));  
    points->push_back(my_red);
    
    }
  
  
  for (unsigned int i=0;i<points->size();i++){
  }
  return 1;
}



double DISPredictor::SigmaInteg(double xmin, double xmax, double q2min, double q2max){

  std::cout << "Calculating total cross section for x, Q2: " << xmin  << "-" <<  xmax << "," << q2min << "-" << q2max << std::endl; 

  double res, err;

  double xl[2] = {q2min,xmin};
  double xu[2] = {q2max,xmax};

  const gsl_rng_type *T;
  gsl_rng *r;

  gsl_monte_function G = { &g, 2, 0 };

  size_t calls = ControlCards::Instance()->fetchValueInt("VegasCalls");
  gsl_rng_env_setup ();

  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  {
    gsl_monte_vegas_state *s = gsl_monte_vegas_alloc (2);
    
    gsl_monte_vegas_integrate (&G, xl, xu, 2, std::min(10000,int(calls/2)), r, s,
                               &res, &err);
    do
      {
        gsl_monte_vegas_integrate (&G, xl, xu, 2, calls, r, s,
                                   &res, &err);
	//	printf ("% .8f pb sigma = % .8f pb "
	//		"chisq/dof = %.1f  \n", res, err, s->chisq);
      }
    while (fabs (s->chisq - 1.0) > 0.5);

    //    printf ("Final result = % .8f pb sigma = % .8f pb "
    //	    "chisq/dof = %.1f  \n", res, err, s->chisq);    
    gsl_monte_vegas_free (s);
  }


  return res;
}

double DISPredictor::SigmaIntegY(double ymin, double ymax, double q2min, double q2max){

  std::cout << "Calculating total cross section for y, Q2: " << ymin  << "-" <<  ymax << "," << q2min << "-" << q2max << std::endl; 

  double res, err;

  double xl[2] = {q2min,ymin};
  double xu[2] = {q2max,ymax};

  const gsl_rng_type *T;
  gsl_rng *r;

  gsl_monte_function G = { &Yg, 2, 0 };


  size_t calls = ControlCards::Instance()->fetchValueInt("VegasCalls");

  gsl_rng_env_setup ();

  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  {
    gsl_monte_vegas_state *s = gsl_monte_vegas_alloc (2);
    
    gsl_monte_vegas_integrate (&G, xl, xu, 2, std::min(10000,int(calls/2)), r, s,
                               &res, &err);
    do
      {
        gsl_monte_vegas_integrate (&G, xl, xu, 2, calls, r, s,
                                   &res, &err);
	//	printf ("% .8f pb sigma = % .8f pb "
	//		"chisq/dof = %.1f  \n", res, err, s->chisq);
      }
    while (fabs (s->chisq - 1.0) > 0.5);

    //    printf ("Final result = % .8f pb sigma = % .8f pb "
	    //	    "chisq/dof = %.1f  \n", res, err, s->chisq);    
    gsl_monte_vegas_free (s);
  }


  return res;
}

double DISPredictor::YSigmaIntegY(double ymin, double ymax, double q2min, double q2max){

  std::cout << "Calculating total cross section for y, Q2: " << ymin  << "-" <<  ymax << "," << q2min << "-" << q2max << std::endl; 

  double res, err;

  double xl[2] = {q2min,ymin};
  double xu[2] = {q2max,ymax};

  const gsl_rng_type *T;
  gsl_rng *r;

  gsl_monte_function G = { &YgY, 2, 0 };


  size_t calls = ControlCards::Instance()->fetchValueInt("VegasCalls");

  gsl_rng_env_setup ();

  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  {
    gsl_monte_vegas_state *s = gsl_monte_vegas_alloc (2);
    
    gsl_monte_vegas_integrate (&G, xl, xu, 2, std::min(10000,int(calls/2)), r, s,
                               &res, &err);
    do
      {
        gsl_monte_vegas_integrate (&G, xl, xu, 2, calls, r, s,
                                   &res, &err);
	//	printf ("% .8f pb sigma = % .8f pb "
	//		"chisq/dof = %.1f  \n", res, err, s->chisq);
      }
    while (fabs (s->chisq - 1.0) > 0.5);

    //    printf ("Final result = % .8f pb sigma = % .8f pb "
    //	    "chisq/dof = %.1f  \n", res, err, s->chisq);    
    gsl_monte_vegas_free (s);
  }


  return res;
}


double DISPredictor::Q2SigmaInteg(double xmin, double xmax, double q2min, double q2max){

  std::cout << "Calculating mean Q2 times total cross section for x, Q2: " << xmin  << "-" <<  xmax << "," << q2min << "-" << q2max << std::endl; 

  double res, err;

  double xl[2] = {q2min,xmin};
  double xu[2] = {q2max,xmax};

  const gsl_rng_type *T;
  gsl_rng *r;

  gsl_monte_function G = { &gQ2, 2, 0 };

  size_t calls = ControlCards::Instance()->fetchValueInt("VegasCalls");

  gsl_rng_env_setup ();

  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  {
    gsl_monte_vegas_state *s = gsl_monte_vegas_alloc (2);
    
    gsl_monte_vegas_integrate (&G, xl, xu, 2, std::min(10000,int(calls/2)), r, s,
                               &res, &err);
    do
      {
        gsl_monte_vegas_integrate (&G, xl, xu, 2, calls, r, s,
                                   &res, &err);
	//	printf ("% .8f pb sigma = % .8f pb "
	//		"chisq/dof = %.1f  \n", res, err, s->chisq);
      }
    while (fabs (s->chisq - 1.0) > 0.5);

    //    printf ("Final result = % .8f pb sigma = % .8f pb "
    //	    "chisq/dof = %.1f  \n", res, err, s->chisq);    
    gsl_monte_vegas_free (s);
  }


  return res;
}


double DISPredictor::XSigmaInteg(double xmin, double xmax, double q2min, double q2max){

  std::cout << "Calculating mean X times total cross section for x, Q2: " << xmin  << "-" <<  xmax << "," << q2min << "-" << q2max << std::endl; 

  double res, err;

  double xl[2] = {q2min,xmin};
  double xu[2] = {q2max,xmax};

  const gsl_rng_type *T;
  gsl_rng *r;

  gsl_monte_function G = { &gX, 2, 0 };

  size_t calls = ControlCards::Instance()->fetchValueInt("VegasCalls");

  gsl_rng_env_setup ();

  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  {
    gsl_monte_vegas_state *s = gsl_monte_vegas_alloc (2);
    
    gsl_monte_vegas_integrate (&G, xl, xu, 2, std::min(10000,int(calls/2)), r, s,
                               &res, &err);
    do
      {
        gsl_monte_vegas_integrate (&G, xl, xu, 2, calls, r, s,
                                   &res, &err);
	//	printf ("% .8f pb sigma = % .8f pb "
	//		"chisq/dof = %.1f  \n", res, err, s->chisq);
      }
    while (fabs (s->chisq - 1.0) > 0.5);

    //    printf ("Final result = % .8f pb sigma = % .8f pb "
    //	    "chisq/dof = %.1f  \n", res, err, s->chisq);    
    gsl_monte_vegas_free (s);
  }


  return res;
}

double DISPredictor::CalculateDSigmaDQ2(double q2min, double q2max){
  return  SigmaInteg(ControlCards::Instance()->fetchValueDouble("XMin"),ControlCards::Instance()->fetchValueDouble("XMax"), q2min, q2max)/(q2max-q2min);
}


double DISPredictor::CalculateQ2DSigmaDQ2(double q2min, double q2max){
  return  Q2SigmaInteg(ControlCards::Instance()->fetchValueDouble("XMin"),ControlCards::Instance()->fetchValueDouble("XMax"), q2min, q2max)/(q2max-q2min);
}


double DISPredictor::CalculateDSigmaDX(double xmin, double xmax){
  return  SigmaInteg(xmin,xmax,ControlCards::Instance()->fetchValueDouble("Q2Min"),ControlCards::Instance()->fetchValueDouble("Q2Max"))/(xmax-xmin);
}

double DISPredictor::CalculateXDSigmaDX(double xmin, double xmax){
  return  XSigmaInteg(xmin,xmax,ControlCards::Instance()->fetchValueDouble("Q2Min"),ControlCards::Instance()->fetchValueDouble("Q2Max"))/(xmax-xmin);
}



double DISPredictor::CalculateDSigmaDY(double ymin, double ymax){
  return  SigmaIntegY(ymin,ymax,ControlCards::Instance()->fetchValueDouble("Q2Min"),ControlCards::Instance()->fetchValueDouble("Q2Max"))/(ymax-ymin);
}

double DISPredictor::CalculateYDSigmaDY(double ymin, double ymax){
  return  YSigmaIntegY(ymin,ymax,ControlCards::Instance()->fetchValueDouble("Q2Min"),ControlCards::Instance()->fetchValueDouble("Q2Max"))/(ymax-ymin);
}


void DISPredictor::PrintResults(){
  std::cout <<  "And here are the results: " << std::endl;
  _redSigmas->PrintShort();


  
   if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDX").c_str(),"ON") ||!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2").c_str(),"ON") ||  !strcmp(ControlCards::Instance()->fetchValueString("DSigmaDY").c_str(),"ON") ){

    std::cout << "Differential Cross Sections: " << std::endl;

    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2").c_str(),"ON")){

      std::cout << "Ds/DQ2 (pb GeV^{-2}) in bins: " << std::endl;
      
      for (int i=1; i<=_dSigdQ2->GetNbinsX();i++){
	std::cout << _dSigdQ2->GetBinLowEdge(i)  << " " << _dSigdQ2->GetBinLowEdge(i)+_dSigdQ2->GetBinWidth(i) << " " << _dSigdQ2->GetBinContent(i)  <<std::endl;
      }
      

      if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDQ2Points").c_str(),"AUTO")){
	std::cout << "Mean Q^2 (GeV^{-2}): " << std::endl;
	
	for (int i=1; i<=_dSigdQ2->GetNbinsX();i++){
	  std::cout << _dSigdQ2->GetBinLowEdge(i)  << " " << _dSigdQ2->GetBinLowEdge(i)+_dSigdQ2->GetBinWidth(i) << " " << _Q2dSigdQ2->GetBinContent(i)/_dSigdQ2->GetBinContent(i)  <<std::endl;
	}
      }
      std::cout << "Ds/DQ2 (pb GeV^{-2}) at points: " << std::endl;
      for (unsigned int i=0; i<_dSigmaDQ2->size();i++){
	std::cout << _dSigmaDQ2->at(i).Var()  << " " << _dSigmaDQ2->at(i).DiffSigma() <<std::endl;
      }

    }
    
    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDX").c_str(),"ON")){

      std::cout << "Ds/DX (pb) in bins: " << std::endl;
    
      for (int i=1; i<=_dSigdX->GetNbinsX();i++){
	std::cout << _dSigdX->GetBinLowEdge(i)  << " " << _dSigdX->GetBinLowEdge(i)+_dSigdX->GetBinWidth(i) << " " << _dSigdX->GetBinContent(i)  <<std::endl;
      }
      
      if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDXPoints").c_str(),"AUTO")){      
	std::cout << "Mean X: " << std::endl;
	
	for (int i=0; i<_dSigdX->GetNbinsX();i++){
	  std::cout << _dSigdX->GetBinLowEdge(i)  << " " << _dSigdX->GetBinLowEdge(i)+_dSigdX->GetBinWidth(i) << " " <<_XdSigdX->GetBinContent(i)/_dSigdX->GetBinContent(i)  <<std::endl;
	}
      }
      
      std::cout << "Ds/DX (pb) at points: " << std::endl;
      for (unsigned int i=0; i<_dSigmaDX->size();i++){
	std::cout << _dSigmaDX->at(i).Var()  << " " << _dSigmaDX->at(i).DiffSigma() <<std::endl;
      }

    }
    

    
    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDY").c_str(),"ON")){
      std::cout << "Ds/DY (pb) in bins: " << std::endl;
      for (int i=1; i<=_dSigdY->GetNbinsX();i++){
	std::cout << _dSigdY->GetBinLowEdge(i)  << " " << _dSigdY->GetBinLowEdge(i)+_dSigdY->GetBinWidth(i) << " " << _dSigdY->GetBinContent(i)  <<std::endl;
      }
    }
    if (!strcmp(ControlCards::Instance()->fetchValueString("DSigmaDYPoints").c_str(),"AUTO")){      
      std::cout << "Mean Y: " << std::endl;
      
      for (int i=0; i<_dSigdY->GetNbinsY();i++){
	std::cout << _dSigdY->GetBinLowEdge(i)  << " " << _dSigdY->GetBinLowEdge(i)+_dSigdY->GetBinWidth(i) << " " <<_YdSigdY->GetBinContent(i)/_dSigdY->GetBinContent(i)  <<std::endl;
      }
    
    
      std::cout << "Ds/DY (pb) at points: " << std::endl;
      for (unsigned int i=0; i<_dSigmaDY->size();i++){
	std::cout << _dSigmaDY->at(i).Var()  << " " << _dSigmaDY->at(i).DiffSigma() <<std::endl;
      }
      
    }
    
    
   }
}

void DISPredictor::WriteOutput(){
  my_file->Write();
  my_file->Close();
}


